# Copyright 2017 SrMouraSilva
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest.mock import MagicMock

from pluginsmanager.observer.update_type import UpdateType
from pluginsmanager.observer.observer_manager import ObserverManager
from pluginsmanager.observer.observable_list import ObservableList


class BanksManager(object):
    """
    BanksManager manager the banks. In these is possible add banks,
    obtains the banks and register observers for will be notified when
    occurs changes (like added new pedalboard, rename bank, set effect
    param value or state)

    For use details, view Readme.rst example documentation.

    :param list[Bank] banks: Banks that will be added in this. Useful
                             for loads banks previously loaded, like
                             banks persisted and recovered.
    """

    def __init__(self, banks=None):
        self.banks = ObservableList()
        self.banks.observer = self._banks_observer

        banks = [] if banks is None else banks
        self.observer_manager = ObserverManager()

        for bank in banks:
            self.append(bank)

    def __iter__(self):
        """
        Iterates banks of the banksmanager::

            >>> banks_manager = BanksManager()
            >>> for index, bank in enumerate(banks_manager):
            ...     print(index, '-', bank)

        :return: Iterator for banks list
        """
        return self.banks.__iter__()

    def register(self, observer):
        """
        Register an observer for it be notified when occurs changes.

        For more details, see :class:`.UpdatesObserver`

        :param UpdatesObserver observer: Observer that will be notified then occurs changes
        """
        self.observer_manager.append(observer)
        observer.manager = self

    def unregister(self, observer):
        """
        Remove the observers of the observers list.
        It will not receive any more notifications when occurs changes.

        :param UpdatesObserver observer: Observer you will not receive any more notifications then
                                         occurs changes.
        """
        self.observer_manager.observers.remove(observer)
        observer.manager = None

    def append(self, bank):
        """
        Append the bank in banks manager. It will be monitored, changes in this
        will be notified for the notifiers.

        :param Bank bank: Bank that will be added in this
        """
        self.banks.append(bank)

    def _banks_observer(self, update_type, bank, index, **kwargs):
        if update_type == UpdateType.CREATED:
            self._init_bank(bank)

        elif update_type == UpdateType.UPDATED:
            self._init_bank(bank)

            old_bank = kwargs['old']
            if old_bank not in self:
                self._clear_bank(old_bank)

        elif update_type == UpdateType.DELETED:
            self._clear_bank(bank)

        self.observer_manager.on_bank_updated(bank, update_type, index=index, origin=self, **kwargs)

    def _init_bank(self, bank):
        bank.manager = self
        bank.observer = self.observer_manager

    def _clear_bank(self, bank):
        bank.manager = None
        bank.observer_manager = MagicMock()

    def enter_scope(self, observer):
        """
        Informs that changes occurs by the ``observer`` and isn't necessary
        informs the changes for observer

        :param UpdatesObserver observer: Observer that causes changes
        """
        self.observer_manager.enter_scope(observer)

    def exit_scope(self):
        """
        Closes the last observer scope added
        """
        self.observer_manager.exit_scope()

    @property
    def observers(self):
        """
        :return: Observers registered in BanksManager instance
        """
        return self.observer_manager.observers



# Copyright 2017 SrMouraSilva
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABCMeta, abstractmethod


class Lv2PortMixing(object, metaclass=ABCMeta):
    """
    Contains the default implementation of System ports:
    :class:`.Lv2Input`, :class:`.Lv2Output`,
    :class:`.Lv2MidiInput` and :class:`.Lv2MidiInput`
    """

    def __init__(self, *args, **kwargs):
        super(Lv2PortMixing, self).__init__(*args, **kwargs)
        self._data = None

    @property
    @abstractmethod
    def data(self):
        """
        :return dict: Metadata used for provide the required information
                      in this object
        """
        pass

    def __str__(self):
        return self.data['name']

    @property
    def symbol(self):
        return self.data['symbol']

    @property
    def __dict__(self):
        dictionary = super(Lv2PortMixing, self).__dict__
        dictionary['index'] = self.data['index']

        return dictionary



# Copyright 2017 SrMouraSilva
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABCMeta, abstractmethod

from unittest.mock import MagicMock


class ParamError(Exception):

    def __init__(self, message):
        super(ParamError, self).__init__(message)
        self.message = message


class Param(metaclass=ABCMeta):
    """
    :class:`.Param` represents an Audio Plugin Parameter::

        >>> my_awesome_effect
        <Lv2Effect object as 'Calf Reverb' active at 0x7fd58d874ba8>
        >>> my_awesome_effect.params
        (<Lv2Param object as value=1.5 [0.4000000059604645 - 15.0] at 0x7fd587f77908>, <Lv2Param object as value=5000.0 [2000.0 - 20000.0] at 0x7fd587f7a9e8>, <Lv2Param object as value=2 [0 - 5] at 0x7fd587f7cac8>, <Lv2Param object as value=0.5 [0.0 - 1.0] at 0x7fd587f7eba8>, <Lv2Param object as value=0.25 [0.0 - 2.0] at 0x7fd58c576c88>, <Lv2Param object as value=1.0 [0.0 - 2.0] at 0x7fd58c578d68>, <Lv2Param object as value=0.0 [0.0 - 500.0] at 0x7fd58c57ae80>, <Lv2Param object as value=300.0 [20.0 - 20000.0] at 0x7fd58c57df98>, <Lv2Param object as value=5000.0 [20.0 - 20000.0] at 0x7fd58c5810f0>)

        >>> param = my_awesome_effect.params[0]
        >>> param
        <Lv2Param object as value=1.5 [0.4000000059604645 - 15.0] at 0x7fd587f77908>

        >>> param.default
        1.5
        >>> param.value = 14

        >>> symbol = param.symbol
        >>> symbol
        'decay_time'
        >>> param == my_awesome_effect.params[symbol]
        True

    :param Effect effect: Effect in which this parameter belongs
    :param default: Default value (initial value parameter)
    """

    def __init__(self, effect, default):
        self._effect = effect
        self._value = default
        self._default = default

        self.observer = MagicMock()

    @property
    def effect(self):
        """
        :return: Effect in which this parameter belongs
        """
        return self._effect

    @property
    def default(self):
        """
        Default parameter value.
        Then a effect is instanced, the value initial for a parameter is
        your default value.

        :getter: Default parameter value.
        """
        return self._default

    @property
    def value(self):
        """
        Parameter value

        :getter: Current value
        :setter: Set the current value
        """
        return self._value

    @value.setter
    def value(self, new_value):
        if self._value == new_value:
            return

        if not(self.minimum <= new_value <= self.maximum):
            msg = 'New value out of range: {} [{} - {}]'.format(
                new_value,
                self.minimum,
                self.maximum
            )
            raise ParamError(msg)

        self._value = new_value
        self.observer.on_param_value_changed(self)

    @property
    @abstractmethod
    def minimum(self):
        """
        :return: Smaller value that the parameter can assume
        """
        pass

    @property
    @abstractmethod
    def maximum(self):
        """
        :return: Greater value that the parameter can assume
        """
        pass

    @property
    @abstractmethod
    def symbol(self):
        """
        :return: Param identifier
        """
        pass

    def __repr__(self, *args, **kwargs):
        return "<{} object as value={} [{} - {}] at 0x{:x}>".format(
            self.__class__.__name__,
            self.value,
            self.minimum,
            self.maximum,
            id(self)
        )

    @property
    def json(self):
        """
        Get a json decodable representation of this param

        :return dict: json representation
        """
        return self.__dict__

    @property
    def __dict__(self):
        return {
            'index': self.index,
            'symbol': self.symbol,
            'value': self.value,
        }

    @property
    def index(self):
        return self.effect.params.index(self)



# Copyright 2017 SrMouraSilva
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABCMeta, abstractmethod


class UpdatesObserver(metaclass=ABCMeta):
    """
    The :class:`.UpdatesObserver` is an abstract class definition for
    treatment of changes in some class model. Your methods are called
    when occurs any change in :class:`.Bank`, :class:`.Pedalboard`,
    :class:`.Effect`, etc.

    To do this, it is necessary that the :class:`.UpdatesObserver` objects
    be registered in some manager, so that it reports the changes. An
    example of a manager is :class:`.BanksManager`.
    """

    def __init__(self):
        self.manager = None

    def __enter__(self):
        if self.manager is not None:
            self.manager.enter_scope(self)

    def __exit__(self, type, value, traceback):
        if self.manager is not None:
            self.manager.exit_scope()

    @abstractmethod
    def on_bank_updated(self, bank, update_type, index, origin, **kwargs):
        """
        Called when changes occurs in any :class:`.Bank`

        :param Bank bank: Bank changed.
        :param UpdateType update_type: Change type
        :param int index: Bank index (or old index if update_type == UpdateType.DELETED)
        :param BanksManager origin: BanksManager that the bank is (or has) contained
        :param Bank: Contains the old bank occurs a `UpdateType.UPDATED`
        """
        pass

    @abstractmethod
    def on_pedalboard_updated(self, pedalboard, update_type, index, origin, **kwargs):
        """
        Called when changes occurs in any :class:`.Pedalboard`

        :param Pedalboard pedalboard: Pedalboard changed
        :param UpdateType update_type: Change type
        :param int index: Pedalboard index (or old index if update_type == UpdateType.DELETED)
        :param Bank origin: Bank that the pedalboard is (or has) contained
        :param Pedalboard old: Contains the old pedalboard when occurs a `UpdateType.UPDATED`
        """
        pass

    @abstractmethod
    def on_effect_updated(self, effect, update_type, index, origin, **kwargs):
        """
        Called when changes occurs in any :class:`.Effect`

        :param Effect effect: Effect changed
        :param UpdateType update_type: Change type
        :param int index: Effect index (or old index if update_type == UpdateType.DELETED)
        :param Pedalboard origin: Pedalboard that the effect is (or has) contained
        """
        pass

    @abstractmethod
    def on_effect_status_toggled(self, effect, **kwargs):
        """
        Called when any :class:`.Effect` status is toggled

        :param Effect effect: Effect when status has been toggled
        """
        pass

    @abstractmethod
    def on_param_value_changed(self, param, **kwargs):
        """
        Called when a param value change

        :param Param param: Param with value changed
        """
        pass

    @abstractmethod
    def on_connection_updated(self, connection, update_type, pedalboard, **kwargs):
        """
        Called when changes occurs in any :class:`pluginsmanager.model.connection.Connection` of Pedalboard
        (adding, updating or removing connections)

        :param pluginsmanager.model.connection.Connection connection: Connection changed
        :param UpdateType update_type: Change type
        :param Pedalboard pedalboard: Pedalboard that the connection is (or has) contained
        """
        pass



# Copyright 2017 SrMouraSilva
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from os import path
from setuptools import setup


def readme():
    here = path.abspath(path.dirname(__file__))

    with open(path.join(here, 'README.rst'), encoding='UTF-8') as f:
        return f.read()

setup(
    name='PedalPi-PluginsManager',
    version='0.6.0alpha1',

    description='Pythonic management of LV2 audio plugins with mod-host.',
    long_description=readme(),

    url='https://github.com/PedalPi/PluginsManager',

    author='Paulo Mateus Moura da Silva (SrMouraSilva)',
    author_email='mateus.moura@hotmail.com',
    maintainer='Paulo Mateus Moura da Silva (SrMouraSilva)',
    maintainer_email='mateus.moura@hotmail.com',

    license="Apache Software License v2",

    packages=[
        'pluginsmanager',
        'pluginsmanager/jack',

        'pluginsmanager/model',
        'pluginsmanager/model/lv2',
        'pluginsmanager/model/system',

        'pluginsmanager/observer',
        'pluginsmanager/observer/autosaver',
        'pluginsmanager/observer/mod_host',

        'pluginsmanager/util',
        'pluginsmanager/util/builder',
    ],
    package_data={
        'pluginsmanager/model/lv2': ['plugins.json']
    },
    install_requires=[
        'JACK-Client',
        'pyaudio'
    ],

    test_suite='test',
    tests_require=['JACK-Client', 'pytest', 'pytest-cov'],

    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Topic :: Multimedia :: Sound/Audio',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6'
    ],
    keywords='pedal-pi mod-host lv2 audio plugins-manager',

    platforms='Linux',
)



# Copyright 2017 SrMouraSilva
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABCMeta

from pluginsmanager.model.audio_port import AudioPort


class Input(AudioPort, metaclass=ABCMeta):
    """
    Input is the medium in which the audio will go into effect to be processed.

    Effects usually have a one (mono) or two inputs (stereo L + stereo R). But this
    isn't a rule: Some have only :class:`.Output`, like audio frequency generators,
    others have more than two.

    For obtains the inputs::

        >>> my_awesome_effect
        <Lv2Effect object as 'Calf Reverb' active at 0x7fd58d874ba8>
        >>> my_awesome_effect.inputs
        (<Lv2Input object as In L at 0x7fd58c583208>, <Lv2Input object as In R at 0x7fd58c587320>)

        >>> effect_input = my_awesome_effect.inputs[0]
        >>> effect_input
        <Lv2Input object as In L at 0x7fd58c583208>

        >>> symbol = effect_input.symbol
        >>> symbol
        'in_l'

        >>> my_awesome_effect.inputs[symbol] == effect_input
        True

    For connections between effects, see :meth:`~pluginsmanager.model.pedalboard.Pedalboard.connect()`
    and :meth:`~pluginsmanager.model.pedalboard.Pedalboard.disconnect()` :class:`.Pedalboard` class methods.

    :param Effect effect: Effect of input
    """

    @property
    def index(self):
        """
        :return: Input index in the your effect
        """
        return self.effect.inputs.index(self)



# Copyright 2017 SrMouraSilva
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABCMeta, abstractmethod
from unittest.mock import MagicMock

from pluginsmanager.util.dict_tuple import DictTuple


class Effect(metaclass=ABCMeta):
    """
    Representation of a audio plugin instance - LV2 plugin encapsulated as a jack client.

    Effect contains a `active` status (off=bypass), a list of :class:`.Param`,
    a list of :class:`.Input` and a list of :class:`~pluginsmanager.model.connection.Connection`::

        >>> reverb = builder.build('http://calf.sourceforge.net/plugins/Reverb')
        >>> pedalboard.append(reverb)
        >>> reverb
        <Lv2Effect object as 'Calf Reverb' active at 0x7fd58d874ba8>

        >>> reverb.active
        True
        >>> reverb.toggle()
        >>> reverb.active
        False
        >>> reverb.active = True
        >>> reverb.active
        True

        >>> reverb.inputs
        (<Lv2Input object as In L at 0x7fd58c583208>, <Lv2Input object as In R at 0x7fd58c587320>)
        >>> reverb.outputs
        (<Lv2Output object as Out L at 0x7fd58c58a438>, <Lv2Output object as Out R at 0x7fd58c58d550>)
        >>> reverb.params
        (<Lv2Param object as value=1.5 [0.4000000059604645 - 15.0] at 0x7fd587f77908>, <Lv2Param object as value=5000.0 [2000.0 - 20000.0] at 0x7fd587f7a9e8>, <Lv2Param object as value=2 [0 - 5] at 0x7fd587f7cac8>, <Lv2Param object as value=0.5 [0.0 - 1.0] at 0x7fd587f7eba8>, <Lv2Param object as value=0.25 [0.0 - 2.0] at 0x7fd58c576c88>, <Lv2Param object as value=1.0 [0.0 - 2.0] at 0x7fd58c578d68>, <Lv2Param object as value=0.0 [0.0 - 500.0] at 0x7fd58c57ae80>, <Lv2Param object as value=300.0 [20.0 - 20000.0] at 0x7fd58c57df98>, <Lv2Param object as value=5000.0 [20.0 - 20000.0] at 0x7fd58c5810f0>)

    :param Pedalboard pedalboard: Pedalboard where the effect lies.
    """

    def __init__(self):
        self.pedalboard = None
        self._active = True

        self._params = ()
        self._inputs = DictTuple([], lambda: None)
        self._outputs = DictTuple([], lambda: None)
        self._midi_inputs = DictTuple([], lambda: None)
        self._midi_outputs = DictTuple([], lambda: None)

        self._observer = MagicMock()

    @property
    def observer(self):
        return self._observer

    @observer.setter
    def observer(self, observer):
        self._observer = observer

        for param in self.params:
            param.observer = self.observer

    @property
    def params(self):
        """
        :return list[Param]: Params of effect
        """
        return self._params

    @property
    def inputs(self):
        """
        :return list[Input]: Inputs of effect
        """
        return self._inputs

    @property
    def outputs(self):
        """
        :return list[Output]: Outputs of effect
        """
        return self._outputs

    @property
    def midi_inputs(self):
        """
        :return list[MidiInput]: MidiInputs of effect
        """
        return self._midi_inputs

    @property
    def midi_outputs(self):
        """
        :return list[MidiOutput]: MidiOutputs of effect
        """
        return self._midi_outputs

    @property
    def active(self):
        """
        Effect status: active or bypass

        :getter: Current effect status
        :setter: Set the effect Status
        :type: bool
        """
        return self._active

    @active.setter
    def active(self, status):
        if status == self._active:
            return

        self._active = status
        self.observer.on_effect_status_toggled(self)

    def toggle(self):
        """
        Toggle the effect status: ``self.active = not self.active``
        """
        self.active = not self.active

    @property
    def connections(self):
        """
        :return list[Connection]: Connections that this effects is present (with input or output port)
        """
        function = lambda connection: connection.input.effect == self \
                                   or connection.output.effect == self

        return tuple([c for c in self.pedalboard.connections if function(c)])

    @property
    def json(self):
        """
        Get a json decodable representation of this effect

        :return dict: json representation
        """
        return self.__dict__

    @property
    @abstractmethod
    def __dict__(self):
        pass

    @property
    def index(self):
        """
        Returns the first occurrence of the effect in your pedalboard
        """
        if self.pedalboard is None:
            raise IndexError('Effect not contains a pedalboard')

        return self.pedalboard.effects.index(self)

    @property
    def is_possible_connect_itself(self):
        """
        return bool: Is possible connect the with it self?
        """
        return False

    @property
    def is_unique_for_all_pedalboards(self):
        """
        return bool: Is unique for all pedalboards?
                     Example: :class:`.SystemEffect` is unique for all pedalboards
        """
        return False

    def __repr__(self):
        return "<{} object as '{}' at 0x{:x}>".format(
            self.__class__.__name__,
            str(self),
            id(self)
        )

    @property
    def version(self):
        """
        :return string: Effect version
        """
        return ''



# Copyright 2017 SrMouraSilva
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from glob import glob
from pathlib import Path

from pluginsmanager.observer.autosaver.persistence import Persistence
from pluginsmanager.util.persistence_decoder import PersistenceDecoder


class BanksFiles(object):

    def __init__(self, data_path):
        """
        :param Path data_path: Path that contains the banks
        """
        self.data_path = data_path

    def load(self, system_effect):
        """
        Return a list if banks presents in data_path

        :param SystemEffect system_effect: SystemEffect used in pedalboards
        :return list[Bank]: List with Banks persisted in
                :attr:`~pluginsmanager.observer.autosaver.banks_files.BanksFiles.data_path`
        """
        persistence = PersistenceDecoder(system_effect)

        banks = []

        for file in glob(str(self.data_path) + "/*.json"):
            bank = persistence.read(Persistence.read(file))
            bank._uuid = file.split('/')[-1].split('.json')[0]
            banks.append(bank)

        return banks

    def save(self, banks_manager):
        for bank in banks_manager:
            self.save_bank(bank)

    def save_bank(self, bank):
        """
        Save the bank in your file

        :param Bank bank: Bank that will be persisted
        """
        path = self._bank_path(bank)
        Persistence.save(path, bank.json)

    def delete_bank(self, bank):
        """
        Delete the bank's file

        :param Bank bank: Bank that will be removed
        """
        path = self._bank_path(bank)
        Persistence.delete(path)

    def delete_all_banks(self):
        """
        Delete all banks files.

        Util for manual save, because isn't possible know which banks
        were removed
        """
        for file in glob(str(self.data_path) + "/*.json"):
            Persistence.delete(file)

    def _bank_path(self, bank):
        """
        :param Bank bank: Bank that will be generate your path

        :return string: Bank path .json
        """
        return self.data_path / Path('{}.json'.format(bank.uuid))



import pyaudio


class JackInterfaces(object):
    """
    JackInterface extract informations about the audio interfaces in the system.

    Requires `pyaudio`_. In Debian based systems, install using::

        sudo apt-get install portaudio19-dev python-all-dev
        pip install pyaudio

    .. _pyaudio: https://people.csail.mit.edu/hubert/pyaudio/
    """

    @staticmethod
    def audio_interfaces():
        """
        Extracts audio interfaces data

        :return list[AudioInterface]: Audio interfaces data
        """
        p = pyaudio.PyAudio()

        interfaces = []
        for i in range(p.get_device_count()):
            data = p.get_device_info_by_index(i)
            if 'hw' not in data['name']:
                interfaces.append(AudioInterface(data))

        p.terminate()

        return interfaces


class AudioInterface(object):
    """
    JackInterface contains data about a audio interface

    :param data: PyAudio interface data using `py_audio_instance.get_device_info_by_index(i)`
    """
    def __init__(self, data):
        self._data = data

    @property
    def data(self):
        """
        :return dict: Original PyAudio interface data
        """
        return self._data

    @property
    def name(self):
        """
        :return string: Device name
        """
        return self.data['name'].split('(')[0]

    @property
    def hw(self):
        """
        :return string: String informing hardware card index and device index
                        Example: `'hw:1:0'`
        """
        return self.data['name'].split('(')[1][0:-1]

    @property
    def total_inputs(self):
        """
        :return float: Max input channels in this audio interface
        """
        return self.data['maxInputChannels']

    @property
    def total_outputs(self):
        """
        :return float: Max output channels in this audio interface
        """
        return self.data['maxOutputChannels']

    @property
    def latency(self):
        """
        :return dict: Min and Max default latency for input and output channels
        """
        return {
            'input': {
                'min': self.data['defaultLowInputLatency'],
                'max': self.data['defaultHighInputLatency'],
            },
            'output': {
                'min': self.data['defaultLowOutputLatency'],
                'max': self.data['defaultHighOutputLatency'],
            }
        }

    @property
    def default_sample_rate(self):
        return self.data['defaultSampleRate']

    @property
    def json(self):
        """
        Get a json decodable representation of this JackInterface

        :return dict: json representation
        """
        return self.__dict__

    @property
    def __dict__(self):
        return {
            'name': self.name,
            'hw': self.hw,
            'total_inputs': self.total_inputs,
            'total_outputs': self.total_outputs,
            'latency': self.latency,
            'default_sample_rate': self.default_sample_rate
        }



# Copyright 2017 SrMouraSilva
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import socket


class Connection(object):
    """
    Class responsible for managing an API connection to the mod-host process via socket
    """
    client = None

    def __init__(self, socket_port=5555, address='localhost'):
        self.client = socket.socket()
        self.client.connect((address, socket_port))
        self.client.settimeout(5)

    def send(self, message):
        """
        Sends message to *mod-host*.

        .. note::

            Uses :class:`.ProtocolParser` for a high-level management.
            As example, view :class:`.Host`

        :param string message: Message that will be sent for *mod-host*
        """
        print(message.encode('utf-8'))
        self.client.send(message.encode('utf-8'))
        received = self.client.recv(1024)

        return received

    def close(self):
        """
        Closes socket connection
        """
        self.client.close()



# Copyright 2017 SrMouraSilva
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pluginsmanager.model.effect import Effect
from pluginsmanager.model.system.system_input import SystemInput
from pluginsmanager.model.system.system_output import SystemOutput
from pluginsmanager.model.system.system_midi_input import SystemMidiInput
from pluginsmanager.model.system.system_midi_output import SystemMidiOutput

from pluginsmanager.util.dict_tuple import DictTuple


class SystemEffect(Effect):
    """
    Representation of the system instance (audio cards).

    System output is equivalent with audio input: You connect the
    instrument in the audio card input and it captures and send the
    audio to :class:`.SystemOutput` for you connect in a input plugins.

    System input is equivalent with audio output: The audio card receives
    the audio processed in your :class:`.SystemInput` and send it to audio
    card output for you connects in amplifier, headset.

    Because no autodetection of existing ports in audio card
    has been implemented, you must explicitly inform in the
    creation of the SystemEffect object::

    >>> sys_effect = SystemEffect('system', ('capture_1', 'capture_2'), ('playback_1', 'playback_2'))

    Unlike effects that should be added in the pedalboard, SystemEffects MUST NOT::

    >>> builder = Lv2EffectBuilder()

    >>> pedalboard = Pedalboard('Rocksmith')
    >>> reverb = builder.build('http://calf.sourceforge.net/plugins/Reverb')
    >>> pedalboard.append(reverb)

    However the pedalboard must have the connections::

    >>> pedalboard.connections.append(Connection(sys_effect.outputs[0], reverb.inputs[0]))

    An bypass example::

    >>> pedalboard = Pedalboard('Bypass example')
    >>> sys_effect = SystemEffect('system', ('capture_1', 'capture_2'), ('playback_1', 'playback_2'))
    >>> pedalboard.connections.append(Connection(sys_effect.outputs[0], sys_effect.inputs[0]))
    >>> pedalboard.connections.append(Connection(sys_effect.outputs[1], sys_effect.inputs[1]))

    :param string representation: Audio card representation. Usually 'system'
    :param tuple(string) outputs: Tuple of outputs representation. Usually a output representation
                                  starts with `capture_`
    :param tuple(string) inputs: Tuple of inputs representation. Usually a input representation
                                 starts with `playback_`
    :param tuple(string) midi_outputs: Tuple of midi outputs representation.
    :param tuple(string) midi_inputs: Tuple of midi inputs representation.
    """
    def __init__(self, representation, outputs, inputs, midi_outputs=None, midi_inputs=None):
        super(SystemEffect, self).__init__()

        self.representation = representation

        self._params = tuple()

        inputs = [SystemInput(self, effect_input) for effect_input in inputs]
        self._inputs = DictTuple(inputs, lambda _input: str(_input))

        outputs = [SystemOutput(self, effect_output) for effect_output in outputs]
        self._outputs = DictTuple(outputs, lambda _output: str(_output))

        midi_inputs = midi_inputs if midi_inputs is not None else []
        midi_inputs = [SystemMidiInput(self, effect_input) for effect_input in midi_inputs]
        self._midi_inputs = DictTuple(midi_inputs, lambda _input: str(_input))

        midi_outputs = midi_outputs if midi_outputs is not None else []
        midi_outputs = [SystemMidiOutput(self, effect_output) for effect_output in midi_outputs]
        self._midi_outputs = DictTuple(midi_outputs, lambda _output: str(_output))

    def __str__(self):
        return self.representation

    @property
    def __dict__(self):
        return {
            'technology': 'system',
        }

    @property
    def is_possible_connect_itself(self):
        """
        return bool: Is possible connect the with it self?
        """
        return True

    @property
    def is_unique_for_all_pedalboards(self):
        """
        return bool: Is unique for all pedalboards?
                     Example: :class:`.SystemEffect` is unique for all pedalboards
        """
        return True
